from NooLite_F.MTRF64.MTRF64USBAdapter import MTRF64USBAdapter, IncomingData, OutgoingData, Command, Mode, Action, ResponseCode, IncomingDataException
from NooLite_F.MTRF64.MTRF64Controller import MTRF64Controller




from NooLite_F import NooLiteFController, ModuleInfo, ModuleState, ModuleMode, BrightnessDirection, ModuleType, RemoteControllerListener, BatteryState
from NooLite_F.MTRF64 import MTRF64USBAdapter, IncomingData, Command, Mode, Action, OutgoingData, ResponseCode


class OutgoingDataException(Exception):
    """Base class for response exceptions."""


class ModuleInfoParser(object):
    @staticmethod
    def parse(response: IncomingData) -> ModuleInfo:
        info = None
        if response.command == Command.SEND_STATE:
            info = ModuleInfo()
            info.type = response.data[0]
            info.firmware = response.data[1]
            info.id = response.id

            if response.format == 0:
                info.state = ModuleState(response.data[2] & 0x0F)
                info.mode = ModuleMode(response.data[2] & 0x80)
                info.brightness = response.data[3] / 255

        return info


class MTRF64Controller(NooLiteFController):

    _adapter = None
    _listener_map: {int: RemoteControllerListener} = {}

    _mode_map = {
        ModuleType.NOOLITE: Mode.TX,
        ModuleType.NOOLITE_F: Mode.TX_F,
    }

    def __init__(self, port: str):
        self._adapter = MTRF64USBAdapter(port, self._on_receive)

    # Private
    def _command_mode(self, module_type: ModuleType) -> Mode:
        return self._mode_map[module_type]

    def _send_module_command(self, module_id, channel: int, command: Command, broadcast, mode: Mode, command_data: bytearray = None, fmt: int = None) -> [(bool, ModuleInfo)]:
        data = OutgoingData()

        data.mode = mode
        data.command = command

        if module_id is not None:
            data.id = module_id
            if channel is None:
                data.action = Action.SEND_COMMAND_TO_ID
            else:
                data.channel = channel
                data.action = Action.SEND_COMMAND_TO_ID_IN_CHANNEL
        elif channel is not None:
            data.channel = channel
            if broadcast and mode == Mode.TX_F: # Broadcast used only for NOOLITE-F mode
                data.action = Action.SEND_BROADCAST_COMMAND
            else:
                data.action = Action.SEND_COMMAND
        else:
            raise OutgoingDataException("Module_id and channel are not specified. You must specify at least one of them.")

        if command_data is not None:
            data.data = command_data

        if fmt is not None:
            data.format = fmt

        response = self._adapter.send(data)
        return self._handle_command_responses(response)

    @staticmethod
    def _handle_command_responses(responses: [IncomingData]) -> [(bool, ModuleInfo)]:
        results = []
        for response in responses:
            info = ModuleInfoParser.parse(response)
            status = response.status == ResponseCode.SUCCESS or response.status == ResponseCode.BIND_SUCCESS
            results.append((status, info))
        return results

    @staticmethod
    def _convert_color_brightness(bright: float) -> int:
        if bright >= 1:
            value = 255
        elif bright <= 0:
            value = 0
        else:
            value = int(255 * bright)
        return value

    # Commands
    def off(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.OFF, broadcast, self._command_mode(module_type))

    def on(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.ON, broadcast, self._command_mode(module_type))

    def temporary_on(self, duration: int, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        data: bytearray = bytearray(2)
        data[0] = duration & 0x00FF
        data[1] = duration & 0xFF00
        return self._send_module_command(module_id, channel, Command.TEMPORARY_ON, broadcast, self._command_mode(module_type), data, 6)

    def enable_temporary_on(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        data: bytearray = bytearray(1)
        data[0] = 0
        return self._send_module_command(module_id, channel, Command.MODES, broadcast, self._command_mode(module_type), data, 1)

    def disable_temporary_on(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        data: bytearray = bytearray(1)
        data[0] = 1
        return self._send_module_command(module_id, channel, Command.MODES, broadcast, self._command_mode(module_type), data, 1)

    def switch(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.SWITCH, broadcast, self._command_mode(module_type))

    def brightness_tune(self, direction: BrightnessDirection, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        if direction == BrightnessDirection.UP:
            command = Command.BRIGHT_UP
        else:
            command = Command.BRIGHT_DOWN
        return self._send_module_command(module_id, channel, command, broadcast, self._command_mode(module_type))

    def brightness_tune_back(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.BRIGHT_BACK, broadcast, self._command_mode(module_type))

    def brightness_tune_stop(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.STOP_BRIGHT, broadcast, self._command_mode(module_type))

    def brightness_tune_custom(self, direction: BrightnessDirection, speed: float, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        if speed >= 1:
            value = 127
        elif speed <= 0:
            value = 0
        else:
            value = int(speed * 127)

        if direction == BrightnessDirection.DOWN:
            value = -value - 1

        data: bytearray = bytearray(1)
        data[0] = value & 0xFF

        return self._send_module_command(module_id, channel, Command.BRIGHT_REG, broadcast, self._command_mode(module_type), data, 1)

    def brightness_tune_step(self, direction: BrightnessDirection, step: int = None, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        data: bytearray = None
        fmt: int = None

        if step is not None:
            fmt = 1
            data: bytearray = bytearray(1)
            data[0] = step

        if direction == BrightnessDirection.UP:
            command = Command.BRIGHT_STEP_UP
        else:
            command = Command.BRIGHT_STEP_DOWN

        return self._send_module_command(module_id, channel, command, broadcast, self._command_mode(module_type), data, fmt)

    def set_brightness(self, brightness: float, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        if brightness >= 1:
            value = 255
        elif brightness <= 0:
            value = 0
        else:
            value = 35 + int(120 * brightness)

        print(value)

        data: bytearray = bytearray(1)
        data[0] = value

        return self._send_module_command(module_id, channel, Command.SET_BRIGHTNESS, broadcast, self._command_mode(module_type), data, 1)

    def roll_rgb_color(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.ROLL_COLOR, broadcast, self._command_mode(module_type))

    def switch_rgb_color(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.SWITCH_COLOR, broadcast, self._command_mode(module_type))

    def switch_rgb_mode(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.SWITCH_MODE, broadcast, self._command_mode(module_type))

    def switch_rgb_mode_speed(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.SPEED_MODE, broadcast, self._command_mode(module_type))

    def set_rgb_brightness(self, red: float, green: float, blue: float, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        data: bytearray = bytearray(3)
        data[0] = self._convert_color_brightness(red)
        data[1] = self._convert_color_brightness(green)
        data[2] = self._convert_color_brightness(blue)
        return self._send_module_command(module_id, channel, Command.SET_BRIGHTNESS, broadcast, self._command_mode(module_type), data, 3)

    def load_preset(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.LOAD_PRESET, broadcast, self._command_mode(module_type))

    def save_preset(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.SAVE_PRESET, broadcast, self._command_mode(module_type))

    def read_state(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.READ_STATE, broadcast, self._command_mode(module_type))

    def bind(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.BIND, broadcast, self._command_mode(module_type))

    def unbind(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.UNBIND, broadcast, self._command_mode(module_type))

    def service_mode_on(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        data: bytearray = bytearray(1)
        data[0] = 1
        return self._send_module_command(module_id, channel, Command.SERVICE, broadcast, self._command_mode(module_type), data)

    def service_mode_off(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        return self._send_module_command(module_id, channel, Command.SERVICE, broadcast, self._command_mode(module_type))

    def set_listener(self, channel: int, listener: RemoteControllerListener):
        self._listener_map[channel] = listener
        print(self._listener_map)

    # Listeners
    def _on_receive(self, incoming_data: IncomingData):
        listener: RemoteControllerListener = self._listener_map.get(incoming_data.channel, None)

        if listener is not None:

            if incoming_data.command == Command.ON:
                listener.on_on()
            elif incoming_data.command == Command.OFF:
                listener.on_off()
            elif incoming_data.command == Command.SWITCH:
                listener.on_switch()
            elif incoming_data.command == Command.TEMPORARY_ON:
                delay = None
                if incoming_data.format == 5:
                    delay = incoming_data.data[0]
                elif incoming_data.format == 6:
                    delay = incoming_data.data[0] + (incoming_data.data[1] << 15)
                listener.on_temporary_on(delay)
            elif incoming_data.command == Command.BRIGHT_UP:
                listener.on_brightness_tune(BrightnessDirection.UP)
            elif incoming_data.command == Command.BRIGHT_DOWN:
                listener.on_brightness_tune(BrightnessDirection.DOWN)
            elif incoming_data.command == Command.BRIGHT_BACK:
                listener.on_brightness_tune_back()
            elif incoming_data.command == Command.BRIGHT_STEP_UP:
                if incoming_data.format == 1:
                    step = incoming_data.data[0]
                else:
                    step = None
                listener.on_brightness_tune_step(BrightnessDirection.UP, step)
            elif incoming_data.command == Command.BRIGHT_STEP_DOWN:
                if incoming_data.format == 1:
                    step = incoming_data.data[0]
                else:
                    step = None
                listener.on_brightness_tune_step(BrightnessDirection.DOWN, step)
            elif incoming_data.command == Command.STOP_BRIGHT:
                listener.on_brightness_tune_stop()
            elif incoming_data.command == Command.SET_BRIGHTNESS:
                if incoming_data.format == 3:
                    red = incoming_data.data[0] / 255
                    green = incoming_data.data[1] / 255
                    blue = incoming_data.data[2] / 255
                    listener.on_set_rgb_brightness(red, green, blue)
                elif incoming_data.format == 1:
                    level = (incoming_data.data[0] - 35) / 120
                    if level < 0:
                        level = 0
                    elif level > 1:
                        level = 1
                    listener.on_set_brightness(level)
            elif incoming_data.command == Command.LOAD_PRESET:
                listener.on_load_preset()
            elif incoming_data.command == Command.SAVE_PRESET:
                listener.on_save_preset()
            elif incoming_data.command == Command.ROLL_COLOR:
                listener.on_roll_rgb_color()
            elif incoming_data.command == Command.SWITCH_COLOR:
                listener.on_switch_rgb_color()
            elif incoming_data.command == Command.SWITCH_MODE:
                listener.on_switch_rgb_mode()
            elif incoming_data.command == Command.SPEED_MODE:
                listener.on_switch_rgb_mode_speed()
            elif incoming_data.command == Command.BRIGHT_REG:
                if incoming_data.format == 1:
                    if incoming_data.data[0] & 0x80 == 0x80:
                        direction = BrightnessDirection.UP
                    else:
                        direction = BrightnessDirection.DOWN
                    speed = (incoming_data.data[0] & 0x7F) / 127
                    listener.on_brightness_tune_custom(direction, speed)
            elif incoming_data.command == Command.SENS_TEMP_HUMI:
                # really from PT111 I get fmt = 7, but in specs is specify that fmt should be 3

                if incoming_data.format == 7:

                    battery_bit = (incoming_data.data[1] & 0x80) >> 7
                    if battery_bit:
                        battery = BatteryState.LOW
                    else:
                        battery = BatteryState.OK

                    temp_low = incoming_data.data[0]
                    temp_hi = incoming_data.data[1] & 0x0F
                    temp = (temp_hi << 8) + temp_low
                    if temp > 0x0800:
                        temp = -(0x1000 - temp)
                    temp = temp / 10

                    device_type = (incoming_data.data[1] & 0x70) >> 4
                    if device_type == 2:
                        humi = incoming_data.data[2]
                    else:
                        humi = None

                    analog = incoming_data.data[3] / 255

                    listener.on_temp_humi(temp, humi, battery, analog)
            elif incoming_data.command == Command.BATTERY_LOW:
                    listener.on_battery_low()






from abc import ABC, abstractmethod
from enum import IntEnum


class ModuleType(IntEnum):
    NOOLITE = 0
    NOOLITE_F = 1


class ModuleState(IntEnum):
    OFF = 0,
    ON = 1,
    TEMPORARY_ON = 2


class ModuleMode(IntEnum):
    BIND_OFF = 0,
    BIND_ON = 1,


class BrightnessDirection(IntEnum):
    UP = 0,
    DOWN = 1,


class BatteryState(IntEnum):
    OK = 0,
    LOW = 1,


class ModuleInfo(object):
    state: ModuleState = None
    mode: ModuleMode = None
    brightness: float = None
    id: int = None
    firmware: int = None
    type: int = None

    def __repr__(self):
        return "<ModuleInfo (0x{0:x}), id: 0x{1:x}, type: {2}, firmware: {3}, state: {4}, brightness: {5}, mode: {6}>"\
            .format(id(self), self.id, self.type, self.firmware, self.state, self.brightness, self.mode)


class RemoteControllerListener(ABC):

    def on_on(self):
        pass

    def on_off(self):
        pass

    def on_switch(self):
        pass

    def on_load_preset(self):
        pass

    def on_save_preset(self):
        pass

    def on_temporary_on(self, duration: int):
        pass

    def on_brightness_tune(self, direction: BrightnessDirection):
        pass

    def on_brightness_tune_back(self):
        pass

    def on_brightness_tune_stop(self):
        pass

    def on_brightness_tune_custom(self, direction: BrightnessDirection, speed: float):
        pass

    def on_brightness_tune_step(self, direction: BrightnessDirection, step: int = None):
        pass

    def on_set_brightness(self, brightness: float):
        pass

    def on_roll_rgb_color(self):
        pass

    def on_switch_rgb_color(self):
        pass

    def on_switch_rgb_mode(self):
        pass

    def on_switch_rgb_mode_speed(self):
        pass

    def on_set_rgb_brightness(self, red: float, green: float, blue: float):
        pass

    def on_temp_humi(self, temp: float, humi: int, battery: BatteryState, analog: float):
        pass

    def on_battery_low(self):
        pass


class NooLiteFController(ABC):

    # Base power control
    @abstractmethod
    def off(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Turn off the modules

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def on(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Turn on the modules

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def temporary_on(self, duration: int, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Turn on the modules for a specified time interval

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param duration: the time during which the modules will be turned on, duration measurement equals 5 sec.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def enable_temporary_on(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Enable "temporary on" mode

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def disable_temporary_on(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Disable "temporary on" mode

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def switch(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Switch modules mode (on/off)

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def brightness_tune(self, direction: BrightnessDirection, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Start to increase/decrease brightness

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param direction: direction of the brightness changing
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def brightness_tune_back(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Invert direction of the brightness change

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def brightness_tune_stop(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Stop brightness changing

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def brightness_tune_custom(self, direction: BrightnessDirection, speed: float, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Start to increase/decrease brightness with a specified speed

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param direction: direction of the brightness changing
        :param speed: speed of the brightness changing. The range of value is 0 .. 1.0
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def brightness_tune_step(self, direction: BrightnessDirection, step: int = None, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Increase/decrease brightness once with a specified step

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param direction: direction of the brightness changing
        :param step: step in microseconds. If specify then can have values in range (1..255) or 0 (it is means 256), by default step equals 64
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def set_brightness(self, brightness: float, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Set brightness

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param brightness: brightness level. The range of value is 0 .. 1.0
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def roll_rgb_color(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Start color changing (only for RGB Led modules)

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def switch_rgb_color(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Switch color (only for RGB Led modules)

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def switch_rgb_mode(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Switch color changing modes (only for RGB Led modules)

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def switch_rgb_mode_speed(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Switch speed of the color changing (only for RGB Led modules)

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def set_rgb_brightness(self, red: float, green: float, blue: float, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Set brightness for each rgb color (only for RGB Led modules)

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param red: red color brightness level. The range of value is 0 .. 1.0
        :param green: green color brightness level. The range of value is 0 .. 1.0
        :param blue: blue color brightness level. The range of value is 0 .. 1.0
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def load_preset(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Load saved module state from preset

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def save_preset(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """ Save current module state as preset

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def read_state(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """  Read module state (only for NooLite-F modules)

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def bind(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """  Send bind command to module

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def unbind(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """  Send unbind command to module

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def service_mode_on(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """  Turn on the service mode on module (only for NooLite-F modules)

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def service_mode_off(self, module_id: int = None, channel: int = None, broadcast: bool = False, module_type: ModuleType = ModuleType.NOOLITE_F) -> [(bool, ModuleInfo)]:
        """  Turn off the service mode on module (only for NooLite-F modules)

        :param module_id: the module id. The command will be send to module with specified id (used only for NOOLITE-F modules).
        :param channel: the number of the channel. The command will be send to all modules that are binded with selected channel. If module_id is also specified then command will be send only to appropriate device in channel.
        :param broadcast: broadcast mode. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False). If module_id is specified or mode is NOOLITE then broadcast parameter will be ignored.
        :param module_type: type of the module, used to determine adapter mode for send command (default - NOOLITE_F).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        pass

    @abstractmethod
    def set_listener(self, channel: int, listener: RemoteControllerListener):
        """ Set the remote controls listener. To remove listener pass None to listener param.

        :param channel: channel to which the listener will be assigned
        :param listener: listener
        """
        pass



from enum import IntEnum
from serial import Serial
from struct import Struct
from time import sleep

from threading import *
from queue import Queue, Empty


class Command(IntEnum):
    OFF = 0,
    BRIGHT_DOWN = 1,
    ON = 2,
    BRIGHT_UP = 3,
    SWITCH = 4,
    BRIGHT_BACK = 5,
    SET_BRIGHTNESS = 6,
    LOAD_PRESET = 7,
    SAVE_PRESET = 8,
    UNBIND = 9,
    STOP_BRIGHT = 10,
    BRIGHT_STEP_DOWN = 11,
    BRIGHT_STEP_UP = 12,
    BRIGHT_REG = 13,
    BIND = 15,
    ROLL_COLOR = 16,
    SWITCH_COLOR = 17,
    SWITCH_MODE = 18,
    SPEED_MODE = 19,
    BATTERY_LOW = 20,
    SENS_TEMP_HUMI = 21,
    TEMPORARY_ON = 25,
    MODES = 26,
    READ_STATE = 128,
    WRITE_STATE = 129,
    SEND_STATE = 130,
    SERVICE = 131,
    CLEAR_MEMORY = 132


class Mode(IntEnum):
    TX = 0,
    RX = 1,
    TX_F = 2,
    RX_F = 3,
    SERVICE = 4,
    FIRMWARE_UPDATE = 5


class ResponseCode(IntEnum):
    SUCCESS = 0,
    NO_RESPONSE = 1,
    ERROR = 2,
    BIND_SUCCESS = 3


class Action(IntEnum):
    SEND_COMMAND = 0,
    SEND_BROADCAST_COMMAND = 1,
    READ_RESPONSE = 2,
    BIND_MODE_ON = 3,
    BIND_MODE_OFF = 4,
    CLEAR_CHANNEL = 5,
    CLEAR_MEMORY = 6,
    UNBIND_ADDRESS_FROM_CHANNEL = 7,
    SEND_COMMAND_TO_ID_IN_CHANNEL = 8,
    SEND_COMMAND_TO_ID = 9


class IncomingDataException(Exception):
    """Base class for response exceptions."""


class OutgoingData(object):
    mode: Mode = Mode.TX
    action: Action = Action.SEND_COMMAND
    channel: int = 0
    command: Command = Command.OFF
    format: int = 0
    data: bytearray = bytearray(4)
    id: int = 0

    def __repr__(self):
        return "<Request (0x{0:x}), mode: {1}, action: {2}, channel: {3:d}, command: {4:d}, format: {5:d}, data: {6}, id: 0x{7:x}>"\
            .format(id(self), self.mode, self.action, self.channel, self.command, self.format, self.data, self.id)


class IncomingData(object):
    mode: Mode = None
    status: ResponseCode = None
    channel: int = None
    command: Command = None
    count: int = None
    format: int = None
    data: bytearray = None
    id: int = None

    def __repr__(self):
        return "<Response (0x{0:x}), mode: {1}, status: {2}, packet_count: {3} channel: {4:d}, command: {5:d}, format: {6:d}, data: {7}, id: 0x{8:x}>"\
            .format(id(self), self.mode, self.status, self.count, self.channel, self.command, self.format, self.data, self.id)


class MTRF64USBAdapter(object):
    _packet_size = 17
    _serial = None
    _read_thread = None
    _command_response_queue = Queue()
    _incoming_queue = Queue()
    _listener_thread = None
    _listener = None

    def __init__(self, port: str, on_receive_data=None):
        self._serial = Serial(baudrate=9600)
        self._serial.port = port
        self._serial.open()

        self._listener = on_receive_data

        self._read_thread = Thread(target=self._read_loop)
        self._read_thread.daemon = True
        self._read_thread.start()

        self._listener_thread = Thread(target=self._read_from_incoming_queue)
        self._listener_thread.daemon = True
        self._listener_thread.start()

    def send(self, data: OutgoingData) -> [IncomingData]:
        responses = []

        packet = self._build(data)
        print("Send:\n - request: {0},\n - packet: {1}".format(data, packet))

        with self._command_response_queue.mutex:
            self._command_response_queue.queue.clear()
        self._serial.write(packet)

        try:
            while True:
                response = self._command_response_queue.get(timeout=1)
                responses.append(response)
                if response.count == 0:
                    break

        except Empty as err:
            print("Error receiving response: {0}".format(err))

        # For NooLite.TX we should make a bit delay. Adapter send the response without waiting until command was delivered.
        # So if we send new command until previous command was sent to module, adapter will ignore new command. Note:
        if data.mode == Mode.TX or data.mode == Mode.RX:
            sleep(0.05)

        return responses

    # Private
    def _crc(self, data) -> int:
        sum = 0
        for i in range(0, len(data)):
            sum = sum + data[i]
        sum = sum & 0xFF
        return sum

    def _build(self, data: OutgoingData) -> bytes:
        format_begin = Struct(">BBBBBBB4sI")
        format_end = Struct("BB")

        packet = format_begin.pack(171, data.mode, data.action, 0, data.channel, data.command, data.format, data.data, data.id)
        packet_end = format_end.pack(self._crc(packet), 172)

        packet = packet + packet_end

        return packet

    def _parse(self, packet: bytes) -> IncomingData:
        if len(packet) != self._packet_size:
            raise IncomingDataException("Invalid packet size: {0}".format(len(packet)))

        format = Struct(">BBBBBBB4sIBB")

        data = IncomingData()
        start_byte, data.mode, data.status, data.count, data.channel, data.command, data.format, data.data, data.id, crc, stop_byte = format.unpack(packet)

        if (start_byte != 173) or (stop_byte != 174) or (crc != self._crc(packet[0:-2])):
            raise IncomingDataException("Invalid response")

        return data

    def _read_loop(self):
        while True:
            packet = self._serial.read(self._packet_size)
            try:
                data = self._parse(packet)
                print("Receive:\n - packet: {0},\n - data: {1}".format(packet, data))

                if data.mode == Mode.TX or data.mode == Mode.TX_F:
                    self._command_response_queue.put(data)
                elif data.mode == Mode.RX or data.mode == Mode.RX_F:
                    self._incoming_queue.put(data)
                else:
                    pass

            except IncomingDataException as err:
                print("Packet error: {0}".format(err))
                pass

    def _read_from_incoming_queue(self):
        while True:
            input_data = self._incoming_queue.get()
            if self._listener is not None:
                self._listener(input_data)



from setuptools import setup
import io

with io.open('README.rst', encoding="utf-8") as readme_file:
    long_description = readme_file.read()

setup(
    name="NooLite_F",
    packages=["NooLite_F", "NooLite_F.MTRF64"],
    version="0.0.9",
    license="MIT License",
    description="Module to work with NooLite/NooLite-F modules via MTRF-64-USB adapter",
    long_description=long_description,
    author="Sergey Prytkov",
    author_email="sergej.prytkov@gmail.com",
    url="https://github.com/SergejPr/NooLite-F",
    keywords="noolite noolite-f noolitef",
    install_requires=["pyserial"],
    platforms="any",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Home Automation",
        "Topic :: System :: Hardware",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
    ]
)



from NooLite_F import NooLiteFController, ModuleType, ModuleInfo, BrightnessDirection


class Switch(object):

    def __init__(self, controller: NooLiteFController, module_id: int = None, channel: int = None, module_type: ModuleType = ModuleType.NOOLITE_F, broadcast_mode: bool = False):
        self._channel = channel
        self._module_type = module_type
        self._broadcast_mode = broadcast_mode
        self._controller = controller
        self._module_id = module_id

    def on(self) -> [(bool, ModuleInfo)]:
        return self._controller.on(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def off(self) -> [(bool, ModuleInfo)]:
        return self._controller.off(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def switch(self) -> [(bool, ModuleInfo)]:
        return self._controller.switch(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def load_preset(self) -> [(bool, ModuleInfo)]:
        return self._controller.load_preset(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def save_preset(self) -> [(bool, ModuleInfo)]:
        return self._controller.save_preset(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def read_state(self) -> [(bool, ModuleInfo)]:
        return self._controller.read_state(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def bind(self) -> [(bool, ModuleInfo)]:
        return self._controller.bind(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def unbind(self) -> [(bool, ModuleInfo)]:
        return self._controller.bind(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def service_mode_on(self) -> [(bool, ModuleInfo)]:
        return self._controller.service_mode_on(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def service_mode_off(self) -> [(bool, ModuleInfo)]:
        return self._controller.service_mode_off(self._module_id, self._channel, self._broadcast_mode, self._module_type)


class ExtendedSwitch(Switch):

    def temporary_on(self, duration: int) -> [(bool, ModuleInfo)]:
        return self._controller.temporary_on(duration, self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def enable_temporary_on(self) -> [(bool, ModuleInfo)]:
        return self._controller.enable_temporary_on(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def disable_temporary_on(self) -> [(bool, ModuleInfo)]:
        return self._controller.disable_temporary_on(self._module_id, self._channel, self._broadcast_mode, self._module_type)


class Dimmer(ExtendedSwitch):

    def brightness_tune(self, direction: BrightnessDirection) -> [(bool, ModuleInfo)]:
        return self._controller.brightness_tune(direction, self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def brightness_tune_back(self) -> [(bool, ModuleInfo)]:
        return self._controller.brightness_tune_back(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def brightness_tune_stop(self) -> [(bool, ModuleInfo)]:
        return self._controller.brightness_tune_stop(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def brightness_tune_custom(self, direction: BrightnessDirection, speed: float):
        return self._controller.brightness_tune_custom(direction, speed, self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def brightness_tune_step(self, direction: BrightnessDirection, step: int = None):
        return self._controller.brightness_tune_step(direction, step, self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def set_brightness(self, brightness: float) -> [(bool, ModuleInfo)]:
        return self._controller.set_brightness(brightness, self._module_id, self._channel, self._broadcast_mode, self._module_type)


class RGBLed(Switch):

    def brightness_tune(self, direction: BrightnessDirection) -> [(bool, ModuleInfo)]:
        return self._controller.brightness_tune(direction, self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def brightness_tune_back(self) -> [(bool, ModuleInfo)]:
        return self._controller.brightness_tune_back(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def brightness_tune_stop(self) -> [(bool, ModuleInfo)]:
        return self._controller.brightness_tune_stop(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def set_brightness(self, brightness: float) -> [(bool, ModuleInfo)]:
        return self._controller.set_brightness(brightness, self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def roll_rgb_color(self) -> [(bool, ModuleInfo)]:
        return self._controller.roll_rgb_color(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def switch_rgb_color(self) -> [(bool, ModuleInfo)]:
        return self._controller.switch_rgb_color(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def switch_rgb_mode(self) -> [(bool, ModuleInfo)]:
        return self._controller.switch_rgb_mode(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def switch_rgb_mode_speed(self) -> [(bool, ModuleInfo)]:
        return self._controller.switch_rgb_mode_speed(self._module_id, self._channel, self._broadcast_mode, self._module_type)

    def set_rgb_brightness(self, red: float, green: float, blue: float) -> [(bool, ModuleInfo)]:
        return self._controller.set_rgb_brightness(red, green, blue, self._module_id, self._channel, self._broadcast_mode, self._module_type)



from NooLite_F.NooLiteFController import NooLiteFController, ModuleType, ModuleState, ModuleMode, ModuleInfo, BrightnessDirection, RemoteControllerListener, BatteryState
from NooLite_F.Modules import Switch, ExtendedSwitch, Dimmer, RGBLed
from NooLite_F.Sensors import Sensor, TempHumiSensor, MovementDetector, RemoteController, RGBRemoteController



from NooLite_F import NooLiteFController, RemoteControllerListener, BatteryState, BrightnessDirection


class Sensor(RemoteControllerListener):

    def __init__(self, controller: NooLiteFController, channel: int, on_battery_low):
        self._controller = controller
        self._channel = channel
        self._battery_low_listener = on_battery_low
        self._controller.set_listener(channel, self)

    def on_battery_low(self):
        if self._battery_low_listener is not None:
            self._battery_low_listener()


class TempHumiSensor(Sensor):

    def __init__(self, controller: NooLiteFController, channel: int, on_temp_humi, on_battery_low=None):
        super().__init__(controller, channel, on_battery_low)
        self._temp_humi_listener = on_temp_humi

    def on_temp_humi(self, temp: float, humi: int, battery: BatteryState, analog: float):
        if self._temp_humi_listener is not None:
            self._temp_humi_listener(temp, humi, battery, analog)


class MovementDetector(Sensor):

    def __init__(self, controller: NooLiteFController, channel: int, on_movement, on_battery_low=None):
        super().__init__(controller, channel, on_battery_low)
        self._movement_listener = on_movement

    def on_temporary_on(self, duration: int):
        if self._movement_listener is not None:
            self._movement_listener()


class RemoteController(Sensor):
    def __init__(self, controller: NooLiteFController, channel: int,
                 on_on=None, on_off=None, on_switch=None,
                 on_tune_start=None, on_tune_back=None, on_tune_stop=None,
                 on_load_preset=None, on_save_preset=None, on_battery_low=None):

        super().__init__(controller, channel, on_battery_low)
        self._on_listener = on_on
        self._off_listener = on_off
        self._switch_listener = on_switch
        self._tune_listener = on_tune_start
        self._back_listener = on_tune_back
        self._stop_listener = on_tune_stop
        self._load_preset_listener = on_load_preset
        self._save_preset_listener = on_save_preset

    def on_on(self):
        if self._on_listener is not None:
            self._on_listener()

    def on_off(self):
        if self._off_listener is not None:
            self._off_listener()

    def on_switch(self):
        if self._switch_listener is not None:
            self._switch_listener()

    def on_load_preset(self):
        if self._load_preset_listener is not None:
            self._load_preset_listener()

    def on_save_preset(self):
        if self._save_preset_listener is not None:
            self._save_preset_listener()

    def on_brightness_tune(self, direction: BrightnessDirection):
        if self._tune_listener is not None:
            self._tune_listener(direction)

    def on_brightness_tune_stop(self):
        if self._stop_listener is not None:
            self._stop_listener()

    def on_brightness_tune_back(self):
        if self._back_listener is not None:
            self._back_listener()


class RGBRemoteController(Sensor):
    def __init__(self, controller: NooLiteFController, channel: int,
                 on_switch=None, on_tune_back=None, on_tune_stop=None,
                 on_roll_color=None, on_switch_color=None, on_switch_mode=None, on_switch_speed=None, on_battery_low=None):

        super().__init__(controller, channel, on_battery_low)
        self._switch_listener = on_switch
        self._back_listener = on_tune_back
        self._stop_listener = on_tune_stop
        self._roll_color_listener = on_roll_color
        self._switch_color_listener = on_switch_color
        self._switch_mode_listener = on_switch_mode
        self._switch_speed_listener = on_switch_speed

    def on_switch(self):
        if self._switch_listener is not None:
            self._switch_listener()

    def on_brightness_tune_stop(self):
        if self._stop_listener is not None:
            self._stop_listener()

    def on_brightness_tune_back(self):
        if self._back_listener is not None:
            self._back_listener()

    def on_switch_rgb_color(self):
        if self._switch_color_listener is not None:
            self._switch_color_listener()

    def on_roll_rgb_color(self):
        if self._roll_color_listener is not None:
            self._roll_color_listener()

    def on_switch_rgb_mode(self):
        if self._switch_mode_listener is not None:
            self._switch_mode_listener()

    def on_switch_rgb_mode_speed(self):
        if self._switch_speed_listener is not None:
            self._switch_speed_listener()
